@@ -60,65 +60,72 @@ static int
 pkg_download(Plisthead *installhead)
 {
 	FILE		*fp;
-	Pkglist  	*pinstall;
+	Pkglist  	*p;
 	struct stat	st;
-	char		pkg_fs[BUFSIZ];
-	char		*p = NULL;
-	off_t		size;
-	int		rc = EXIT_SUCCESS;
+	char		*pkgurl;
+	int		count = 0, i = 1, rc = EXIT_SUCCESS;
 
-	SLIST_FOREACH(pinstall, installhead, next) {
+	/*
+	 * Get total for download counters.
+	 */
+	SLIST_FOREACH(p, installhead, next) {
+		count++;
+	}
+
+	SLIST_FOREACH(p, installhead, next) {
 		/*
 		 * We don't (yet) support resume so start by explicitly
 		 * removing any existing file.  pkgin_install() has already
 		 * checked to see if it's valid, and we know it is not.
 		 */
-		(void) snprintf(pkg_fs, BUFSIZ, "%s/%s%s", pkgin_cache,
-		    pinstall->depend, PKG_EXT);
-		(void) unlink(pkg_fs);
+		(void) unlink(p->ipkg->pkgfs);
 		(void) umask(DEF_UMASK);
 
-		if (strncmp(pinstall->pkgurl, "file:///", 8) == 0) {
+		if (strncmp(p->ipkg->pkgurl, "file:///", 8) == 0) {
 			/*
 			 * If this package repository URL is file:// we can
 			 * just symlink rather than copying.  We do not support
 			 * file:// URLs with a host component.
 			 */
-			p = &pinstall->pkgurl[7];
+			pkgurl = &p->ipkg->pkgurl[7];
 
-			if (stat(p, &st) != 0) {
+			if (stat(pkgurl, &st) != 0) {
 				fprintf(stderr, MSG_PKG_NOT_AVAIL,
-				    pinstall->depend);
+				    p->ipkg->rpkg->full);
 				rc = EXIT_FAILURE;
 				if (check_yesno(DEFAULT_NO) == ANSW_NO)
 					exit(rc);
-				pinstall->file_size = -1;
+				p->ipkg->file_size = -1;
 				continue;
 			}
 
-			if (symlink(p, pkg_fs) < 0)
+			if (symlink(pkgurl, p->ipkg->pkgfs) < 0) {
 				errx(EXIT_FAILURE,
-				    "Failed to create symlink %s", pkg_fs);
+				    "Failed to create symlink %s",
+				    p->ipkg->pkgfs);
+			}
 
-			size = st.st_size;
+			p->ipkg->file_size = st.st_size;
 		} else {
 			/*
 			 * Fetch via HTTP.  download_pkg() handles printing
 			 * errors from various failure modes, so we handle
 			 * cleanup only.
 			 */
-			if ((fp = fopen(pkg_fs, "w")) == NULL)
-				err(EXIT_FAILURE, MSG_ERR_OPEN, pkg_fs);
+			if ((fp = fopen(p->ipkg->pkgfs, "w")) == NULL)
+				err(EXIT_FAILURE, MSG_ERR_OPEN, p->ipkg->pkgfs);
 
-			if ((size = download_pkg(pinstall->pkgurl, fp)) == -1) {
+			p->ipkg->file_size =
+			    download_pkg(p->ipkg->pkgurl, fp, i++, count);
+
+			if (p->ipkg->file_size == -1) {
 				(void) fclose(fp);
-				(void) unlink(pkg_fs);
+				(void) unlink(p->ipkg->pkgfs);
 				rc = EXIT_FAILURE;
 
 				if (check_yesno(DEFAULT_NO) == ANSW_NO)
 					exit(rc);
 
-				pinstall->file_size = -1;
 				continue;
 			}
 
@@ -130,20 +137,21 @@ pkg_download(Plisthead *installhead)
 		 * specified by the server, this checks that it matches what
 		 * is recorded by pkg_summary.
 		 */
-		if (size != pinstall->file_size) {
-			(void) unlink(pkg_fs);
+		if (p->ipkg->file_size != p->ipkg->rpkg->file_size) {
+			(void) unlink(p->ipkg->pkgfs);
 			rc = EXIT_FAILURE;
 
 			(void) fprintf(stderr, "download error: %s size"
-			    " does not match pkg_summary\n", pinstall->depend);
+			    " does not match pkg_summary\n",
+			    p->ipkg->rpkg->full);
 
 			if (check_yesno(DEFAULT_NO) == ANSW_NO)
 				exit(rc);
 
-			pinstall->file_size = -1;
+			p->ipkg->file_size = -1;
 			continue;
 		}
-	} /* download loop */
+	}
 
 	return rc;
 }
@@ -196,17 +204,17 @@ log_tag(const char *fmt, ...)
 	time_t		now;
 
 	now = time(NULL);
-	tim = *(localtime(&now));
+	tim = *(gmtime(&now));
 
 	va_start(ap, fmt);
 	vsnprintf(log_action, BUFSIZ, fmt, ap);
 	va_end(ap);
 
-	(void)strftime(now_date, DATELEN, "%b %d %H:%M:%S", &tim);
+	(void)strftime(now_date, DATELEN, "%FT%TZ", &tim);
 
 	printf("%s", log_action);
 	if (!verbosity) {
-		fprintf(err_fp, "---%s: %s", now_date, log_action);
+		fprintf(err_fp, "%s %s", now_date, log_action);
 		fflush(err_fp);
 	}
 }
@@ -241,6 +249,45 @@ close_pi_log(int output)
 	warn_count = err_count = said = 0;
 }
 
+int
+action_is_install(action_t action)
+{
+	switch (action) {
+	case ACTION_INSTALL:
+	case ACTION_UPGRADE:
+	case ACTION_REFRESH:
+		return 1;
+		break;
+	default:
+		return 0;
+		break;
+	}
+}
+
+int
+action_is_remove(action_t action)
+{
+	switch (action) {
+	case ACTION_REMOVE:
+	case ACTION_SUPERSEDED:
+		return 1;
+		break;
+	default:
+		return 0;
+		break;
+	}
+}
+
+/*
+ * Sometimes code paths use an indirect pointer via ipkg to a Pkglist entry and
+ * others go direct.
+ */
+static Pkglist *
+get_pkglist_ptr(Pkglist *p)
+{
+	return (p->ipkg) ? p->ipkg : p;
+}
+
 /*
  * Execute "pkg_admin rebuild-tree" to rebuild +REQUIRED_BY files after any
  * operation that changes the installed packages.
@@ -267,33 +314,43 @@ rebuild_required_by(void)
 	return rv;
 }
 
-/* package removal */
+/*
+ * Remove a list of packages.  The package list may contain entries that are
+ * not removals.
+ */
 void
 do_pkg_remove(Plisthead *removehead)
 {
-	Pkglist *premove;
+	Pkglist *pkg, *p;
+	int count = 0, i = 1;
+
+	/*
+	 * Get count of packages that are being removed.
+	 */
+	SLIST_FOREACH(pkg, removehead, next) {
+		p = get_pkglist_ptr(pkg);
+		if (action_is_remove(p->action)) {
+			count++;
+		}
+	}
+
+	/*
+	 * Avoid printing logfile stats if there are no packages to remove.
+	 */
+	if (count == 0)
+		return;
 
 	/* send pkg_delete stderr to logfile */
 	open_pi_log();
 
-	SLIST_FOREACH(premove, removehead, next) {
-		/* file not available in the repository */
-		if (premove->file_size == -1)
+	SLIST_FOREACH(pkg, removehead, next) {
+		p = get_pkglist_ptr(pkg);
+		if (!action_is_remove(p->action))
 			continue;
 
-		if (premove->depend == NULL)
-			/* SLIST corruption, badly installed package */
-			continue;
-
-		/* pkg_install cannot be deleted */
-		if (strcmp(premove->depend, "pkg_install") == 0) {
-			printf("not removing pkg_install...\n");
-			continue;
-		}
-
-		log_tag(MSG_REMOVING, premove->depend);
-		if (fexec(pkg_delete, verb_flag("-f"), premove->depend, NULL)
-			!= EXIT_SUCCESS)
+		log_tag("[%d/%d] removing %s...\n", i++, count, p->lpkg->full);
+		if (fexec(pkg_delete, verb_flag("-f"), p->lpkg->full, NULL)
+		    != EXIT_SUCCESS)
 			err_count++;
 	}
 
@@ -311,35 +368,72 @@ do_pkg_remove(Plisthead *removehead)
 static int
 do_pkg_install(Plisthead *installhead)
 {
-	int		rc = EXIT_SUCCESS;
-	Pkglist		*pinstall;
-	char		pkgpath[BUFSIZ];
-	char		*pflags = verb_flag("-DU");
+	Pkglist		*p;
+	int		count = 0, i = 1, rc = EXIT_SUCCESS;
+	const char	*action, *aflags, *iflags, *pflags;
+
+	/*
+	 * Packages specified on the command line are marked as "keep", while
+	 * their dependencies use the -A pkg_add flag to indicate they are
+	 * automatic packages that can be autoremoved when no longer required.
+	 */
+	iflags = (verbosity) ? "-DUv" : "-DU";
+	aflags = (verbosity) ? "-ADUv" : "-ADU";
 
 	/* send pkg_add stderr to logfile */
 	open_pi_log();
 
-	SLIST_FOREACH(pinstall, installhead, next) {
-		/* file not available in the repository */
-		if (pinstall->file_size == -1)
+	/*
+	 * Get total number of packages we'll be operating on.
+	 */
+	SLIST_FOREACH(p, installhead, next) {
+		if (!action_is_install(p->ipkg->action))
+			continue;
+		if (p->ipkg->file_size == -1)
+			continue;
+		count++;
+	}
+
+	SLIST_FOREACH(p, installhead, next) {
+		if (!action_is_install(p->ipkg->action))
 			continue;
 
-		snprintf(pkgpath, BUFSIZ,
-			"%s/%s%s", pkgin_cache, pinstall->depend, PKG_EXT);
+		/* file not available in the repository */
+		if (p->ipkg->file_size == -1)
+			continue;
 
-		switch (pinstall->action) {
-		case TOREFRESH:
-			log_tag("refreshing %s...\n", pinstall->depend);
+		/*
+		 * If package has been marked as keep (explicitly installed),
+		 * or is an upgrade of a local package that was marked as keep,
+		 * then use the non-automatic flags, otherwise use -A.
+		 */
+		if (p->ipkg->keep || (p->ipkg->lpkg && p->ipkg->lpkg->keep))
+			pflags = iflags;
+		else
+			pflags = aflags;
+
+		switch (p->ipkg->action) {
+		case ACTION_REFRESH:
+			action = "refreshing";
 			break;
-		case TOUPGRADE:
-			log_tag("upgrading %s...\n", pinstall->depend);
+		case ACTION_UPGRADE:
+			action = "upgrading";
 			break;
-		case TOINSTALL:
-			log_tag("installing %s...\n", pinstall->depend);
+		case ACTION_INSTALL:
+			action = "installing";
+			break;
+		default:
+			/*
+			 * action_is_install() earlier already excluded all
+			 * other possible cases.
+			 */
 			break;
 		}
 
-		if (fexec(pkg_add, pflags, pkgpath, NULL) == EXIT_FAILURE)
+		log_tag("[%d/%d] %s %s...\n", i++, count, action,
+		    p->ipkg->rpkg->full);
+		if (fexec(pkg_add, pflags, p->ipkg->pkgfs, NULL)
+		    == EXIT_FAILURE)
 			rc = EXIT_FAILURE;
 	}
 
@@ -405,31 +499,96 @@ action_list(char *flatlist, char *str)
 	return newlist;
 }
 
+static char **
+get_sorted_list(Plisthead *pkgs)
+{
+	Pkglist *pkg, *p;
+	size_t i = 0;
+	char **names;
+
+	/* Get number of entries for names allocation */
+	SLIST_FOREACH(pkg, pkgs, next)
+		i++;
+
+	names = xmalloc((i + 1) * sizeof(char *));
+
+	i = 0;
+	SLIST_FOREACH(pkg, pkgs, next) {
+		p = get_pkglist_ptr(pkg);
+		if (p->rpkg)
+			names[i++] = p->rpkg->full;
+		else
+			names[i++] = p->lpkg->full;
+	}
+	names[i] = NULL;
+
+	qsort(names, i, sizeof(char *), sort_pkg_alpha);
+
+	return names;
+}
+
+static char **
+get_sorted_list_by_action(Plisthead *pkgs, action_t action)
+{
+	Pkglist *pkg, *p;
+	size_t i = 0;
+	char **names;
+
+	/* Get number of entries for names allocation */
+	SLIST_FOREACH(pkg, pkgs, next) {
+		p = get_pkglist_ptr(pkg);
+		if (p->action == action)
+			i++;
+	}
+
+	names = xmalloc((i + 1) * sizeof(char *));
+
+	i = 0;
+	SLIST_FOREACH(pkg, pkgs, next) {
+		p = get_pkglist_ptr(pkg);
+		if (p->action == action) {
+			if (p->rpkg)
+				names[i++] = p->rpkg->full;
+			else
+				names[i++] = p->lpkg->full;
+		}
+	}
+	names[i] = NULL;
+
+	qsort(names, i, sizeof(char *), sort_pkg_alpha);
+
+	return names;
+}
+
 #define H_BUF 6
 
 int
 pkgin_install(char **pkgargs, int do_inst, int upgrade)
 {
+	Plistarray	*conflicts;
 	FILE		*fp;
-	int		installnum = 0, upgradenum = 0, removenum = 0;
+	int		installnum = 0, upgradenum = 0;
 	int		refreshnum = 0, downloadnum = 0;
-	int		rc = EXIT_SUCCESS;
+	int		removenum = 0, supersedenum = 0;
+	int		conflictnum = 0;
+	int		argn, rc = EXIT_SUCCESS;
 	int		privsreqd = PRIVS_PKGINDB;
 	uint64_t	free_space;
 	int64_t		file_size = 0, size_pkg = 0;
 	size_t		len;
 	ssize_t		llen;
-	Pkglist		*pkg;
+	Pkglist		*p;
 	Plisthead	*impacthead, *downloadhead = NULL, *installhead = NULL;
-	char		*p;
+	char		**names;
 	char		*toinstall = NULL, *toupgrade = NULL;
 	char		*torefresh = NULL, *todownload = NULL;
+	char		*toremove = NULL, *tosupersede = NULL;
 	char		*unmet_reqs = NULL;
-	char		pkgpath[BUFSIZ], pkgrepo[BUFSIZ], query[BUFSIZ];
+	char		pkgrepo[BUFSIZ], query[BUFSIZ];
 	char		h_psize[H_BUF], h_fsize[H_BUF], h_free[H_BUF];
 	struct		stat st;
 
-	if (SLIST_EMPTY(&r_plisthead)) {
+	if (is_empty_remote_pkglist()) {
 		printf("%s\n", MSG_EMPTY_AVAIL_PKGLIST);
 		return EXIT_FAILURE;
 	}
@@ -440,7 +599,10 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 	if (!have_privs(privsreqd))
 		errx(EXIT_FAILURE, MSG_DONT_HAVE_RIGHTS);
 
-	/* full impact list */
+	/*
+	 * Calculate full impact (list of packages and actions to perform) of
+	 * requested operation.
+	 */
 	if ((impacthead = pkg_impact(pkgargs, &rc)) == NULL) {
 		printf(MSG_NOTHING_TO_DO);
 		return rc;
@@ -448,37 +610,73 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 
 	/* check for required files */
 	if (!pkg_met_reqs(impacthead))
-		SLIST_FOREACH(pkg, impacthead, next)
-			if (pkg->action == UNMET_REQ)
+		SLIST_FOREACH(p, impacthead, next)
+			if (p->action == ACTION_UNMET_REQ)
 				unmet_reqs =
-					action_list(unmet_reqs, pkg->full);
+				    action_list(unmet_reqs, p->rpkg->full);
 
-	/* browse impact tree */
-	SLIST_FOREACH(pkg, impacthead, next) {
-
-		/*
-		 * Packages being removed need no special handling, account
-		 * for them and move to the next package.
-		 */
-		if (pkg->action == TOREMOVE) {
-			removenum++;
+	/*
+	 * Check for conflicts.
+	 *
+	 * XXX: this should really be done during impact so that we have a
+	 * chance to consider alternatives.
+	 */
+	conflicts = init_array(CONFLICTS_HASH_SIZE);
+	get_conflicts(conflicts);
+	SLIST_FOREACH(p, impacthead, next) {
+		if (!action_is_install(p->action))
 			continue;
-		}
+		if (pkg_has_conflicts(p, conflicts))
+			conflictnum++;
+	}
+	free_array(conflicts);
 
-		/* check for conflicts */
-		if (pkg_has_conflicts(pkg))
-			if (!check_yesno(DEFAULT_NO))
-				goto installend;
+	if (conflictnum && !check_yesno(DEFAULT_NO))
+		goto installend;
+
+	/*
+	 * Set up counters.
+	 */
+	SLIST_FOREACH(p, impacthead, next) {
+		switch (p->action) {
+		case ACTION_INSTALL:
+			installnum++;
+			break;
+		case ACTION_UPGRADE:
+			upgradenum++;
+			break;
+		case ACTION_REFRESH:
+			refreshnum++;
+			break;
+		case ACTION_REMOVE:
+			removenum++;
+			break;
+		case ACTION_SUPERSEDED:
+			supersedenum++;
+			break;
+		default:
+			break;
+		}
+	}
+
+	/*
+	 * Set up download URLs and sizes.
+	 */
+	SLIST_FOREACH(p, impacthead, next) {
+		if (!action_is_install(p->action))
+			continue;
 
 		/*
 		 * Retrieve the correct repository for the package and save it,
 		 * this is used later by pkg_download().
 		 */
-		sqlite3_snprintf(BUFSIZ, query, PKG_URL, pkg->full);
+		sqlite3_snprintf(BUFSIZ, query, PKG_URL, p->rpkg->full);
 		if (pkgindb_doquery(query, pdb_get_value, pkgrepo) != 0)
-			errx(EXIT_FAILURE, MSG_PKG_NO_REPO, pkg->full);
+			errx(EXIT_FAILURE, MSG_PKG_NO_REPO, p->rpkg->full);
 
-		pkg->pkgurl = xasprintf("%s/%s%s", pkgrepo, pkg->full,
+		p->pkgfs = xasprintf("%s/%s%s", pkgin_cache, p->rpkg->full,
+		    PKG_EXT);
+		p->pkgurl = xasprintf("%s/%s%s", pkgrepo, p->rpkg->full,
 		    PKG_EXT);
 
 		/*
@@ -486,64 +684,60 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 		 * its size does not match pkg_summary, then mark it to be
 		 * downloaded.
 		 */
-		(void) snprintf(pkgpath, BUFSIZ, "%s/%s%s", pkgin_cache,
-		    pkg->full, PKG_EXT);
-		if (stat(pkgpath, &st) < 0 || st.st_size != pkg->file_size)
-			pkg->download = 1;
+		if (stat(p->pkgfs, &st) < 0 || st.st_size != p->rpkg->file_size)
+			p->download = 1;
 		else {
 			/*
 			 * If the cached package has the correct size, we must
 			 * verify that the BUILD_DATE has not changed, in case
 			 * the sizes happen to be identical.
 			 */
-			p = xasprintf("%s -Q BUILD_DATE %s", pkg_info, pkgpath);
+			char *s;
+			s = xasprintf("%s -Q BUILD_DATE %s", pkg_info, p->pkgfs);
 
-			if ((fp = popen(p, "r")) == NULL)
-				err(EXIT_FAILURE, "Cannot execute '%s'", p);
-			(void) free(p);
+			if ((fp = popen(s, "r")) == NULL)
+				err(EXIT_FAILURE, "Cannot execute '%s'", s);
+			free(s);
 
-			for (p = NULL, len = 0;
-			     (llen = getline(&p, &len, fp)) > 0;
-			     (void) free(p), p = NULL, len = 0) {
-				if (p[llen - 1] == '\n')
-					p[llen - 1] = '\0';
-				if (pkgstrcmp(p, pkg->build_date))
-					pkg->download = 1;
+			for (s = NULL, len = 0;
+			     (llen = getline(&s, &len, fp)) > 0;
+			     free(s), s = NULL, len = 0) {
+				if (s[llen - 1] == '\n')
+					s[llen - 1] = '\0';
+				if (pkgstrcmp(s, p->rpkg->build_date))
+					p->download = 1;
 			}
+			free(s);
 			(void) pclose(fp);
 		}
 
 		/*
 		 * Don't account for download size if using a file:// repo.
 		 */
-		if (pkg->download) {
+		if (p->download) {
 			downloadnum++;
 			if (strncmp(pkgrepo, "file:///", 8) != 0)
-				file_size += pkg->file_size;
+				file_size += p->rpkg->file_size;
 		}
 
-		if (pkg->old_size_pkg > 0)
-			pkg->size_pkg -= pkg->old_size_pkg;
-
-		size_pkg += pkg->size_pkg;
-
-		switch (pkg->action) {
-		case TOREFRESH:
-			refreshnum++;
-			break;
-		case TOUPGRADE:
-			upgradenum++;
-			break;
-		case TOINSTALL:
-			installnum++;
-			break;
-		}
+		if (p->lpkg && p->lpkg->size_pkg > 0)
+			size_pkg += p->rpkg->size_pkg - p->lpkg->size_pkg;
+		else
+			size_pkg += p->rpkg->size_pkg;
 	}
 
+	/*
+	 * Ensure pretty-printed h_psize is always positive, and adjust output
+	 * messages based on size_pkg.
+	 */
 	(void)humanize_number(h_fsize, H_BUF, file_size, "",
 		HN_AUTOSCALE, HN_B | HN_NOSPACE | HN_DECIMAL);
-	(void)humanize_number(h_psize, H_BUF, size_pkg, "",
-		HN_AUTOSCALE, HN_B | HN_NOSPACE | HN_DECIMAL);
+	if (size_pkg < 0)
+		(void)humanize_number(h_psize, H_BUF, 0 - size_pkg, "",
+			HN_AUTOSCALE, HN_B | HN_NOSPACE | HN_DECIMAL);
+	else
+		(void)humanize_number(h_psize, H_BUF, size_pkg, "",
+			HN_AUTOSCALE, HN_B | HN_NOSPACE | HN_DECIMAL);
 
 	/* check disk space */
 	free_space = fs_room(pkgin_cache);
@@ -570,50 +764,89 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 	}
 
 	/*
-	 * Separate package lists according to action.
+	 * Separate package lists according to action and sort alphabetically.
 	 */
 	downloadhead = order_download(impacthead);
-	SLIST_FOREACH(pkg, downloadhead, next) {
-		todownload = action_list(todownload, pkg->depend);
+
+	names = get_sorted_list(downloadhead);
+	for (argn = 0; names[argn] != NULL; argn++) {
+		todownload = action_list(todownload, names[argn]);
 	}
+	free(names);
 
 	installhead = order_install(impacthead);
-	SLIST_FOREACH(pkg, installhead, next) {
-		switch (pkg->action) {
-		case TOREFRESH:
-			torefresh = action_list(torefresh, pkg->depend);
-			break;
-		case TOUPGRADE:
-			toupgrade = action_list(toupgrade, pkg->depend);
-			break;
-		case TOINSTALL:
-			toinstall = action_list(toinstall, pkg->depend);
-			break;
-		}
+
+	names = get_sorted_list_by_action(installhead, ACTION_REFRESH);
+	for (argn = 0; names[argn] != NULL; argn++) {
+		torefresh = action_list(torefresh, names[argn]);
 	}
+	free(names);
+
+	names = get_sorted_list_by_action(installhead, ACTION_UPGRADE);
+	for (argn = 0; names[argn] != NULL; argn++) {
+		toupgrade = action_list(toupgrade, names[argn]);
+	}
+	free(names);
+
+	names = get_sorted_list_by_action(installhead, ACTION_INSTALL);
+	for (argn = 0; names[argn] != NULL; argn++) {
+		toinstall = action_list(toinstall, names[argn]);
+	}
+	free(names);
+
+	names = get_sorted_list_by_action(installhead, ACTION_REMOVE);
+	for (argn = 0; names[argn] != NULL; argn++) {
+		toremove = action_list(toremove, names[argn]);
+	}
+	free(names);
+
+	names = get_sorted_list_by_action(installhead, ACTION_SUPERSEDED);
+	for (argn = 0; names[argn] != NULL; argn++) {
+		tosupersede = action_list(tosupersede, names[argn]);
+	}
+	free(names);
 
 	printf("\n");
 
 	if (do_inst) {
-		if (refreshnum > 0)
-			printf("%d package%s to refresh:\n%s\n\n", refreshnum,
-			    (refreshnum == 1) ? "" : "s", torefresh);
+		if (refreshnum == 1)
+			printf(MSG_ONE_TO_REFRESH, torefresh);
+		else if (refreshnum > 1)
+			printf(MSG_NUM_TO_REFRESH, refreshnum, torefresh);
 
-		if (upgradenum > 0)
-			printf("%d package%s to upgrade:\n%s\n\n", upgradenum,
-			    (upgradenum == 1) ? "" : "s", toupgrade);
+		if (upgradenum == 1)
+			printf(MSG_ONE_TO_UPGRADE, toupgrade);
+		else if (upgradenum > 1)
+			printf(MSG_NUM_TO_UPGRADE, upgradenum, toupgrade);
 
-		if (installnum > 0)
-			printf("%d package%s to install:\n%s\n\n", installnum,
-			    (installnum == 1) ? "" : "s", toinstall);
+		if (installnum == 1)
+			printf(MSG_ONE_TO_INSTALL, toinstall);
+		else if (installnum > 1)
+			printf(MSG_NUM_TO_INSTALL, installnum, toinstall);
 
-		printf("%d to refresh, %d to upgrade, %d to install\n",
-		    refreshnum, upgradenum, installnum);
-		printf("%s to download, %s to install\n", h_fsize, h_psize);
+		if (removenum == 1)
+			printf(MSG_ONE_TO_REMOVE, toremove);
+		else if (removenum > 1)
+			printf(MSG_NUM_TO_REMOVE, removenum, toremove);
+
+		if (supersedenum == 1)
+			printf(MSG_ONE_TO_SUPERSEDE, tosupersede);
+		else if (supersedenum > 1)
+			printf(MSG_NUM_TO_SUPERSEDE, supersedenum, tosupersede);
+
+		printf(MSG_ALL_TO_ACTION, removenum + supersedenum, refreshnum,
+		    upgradenum, installnum);
+
+		if (size_pkg >= 0)
+			printf(MSG_DOWNLOAD_USED, h_fsize, h_psize);
+		else
+			printf(MSG_DOWNLOAD_FREED, h_fsize, h_psize);
 	} else {
-		printf("%d package%s to download:\n%s\n", downloadnum,
-		    (downloadnum == 1) ? "" : "s", todownload);
-		printf("%s to download\n", h_fsize);
+		if (downloadnum == 1)
+			printf(MSG_ONE_TO_DOWNLOAD, todownload);
+		else
+			printf(MSG_NUM_TO_DOWNLOAD, downloadnum, todownload);
+		printf(MSG_DOWNLOAD, h_fsize);
 	}
 
 	if (unmet_reqs != NULL)/* there were unmet requirements */
@@ -635,19 +868,24 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 		if (!do_inst)
 			goto installend;
 
-		SLIST_FOREACH(pkg, downloadhead, next) {
-			if (pkg->file_size != -1)
+		SLIST_FOREACH(p, downloadhead, next) {
+			if (!action_is_install(p->ipkg->action))
 				continue;
 
-			switch (pkg->action) {
-			case TOREFRESH:
-				refreshnum--;
+			if (p->ipkg->file_size != -1)
+				continue;
+
+			switch (p->ipkg->action) {
+			case ACTION_INSTALL:
+				installnum--;
 				break;
-			case TOUPGRADE:
+			case ACTION_UPGRADE:
 				upgradenum--;
 				break;
-			case TOINSTALL:
-				installnum--;
+			case ACTION_REFRESH:
+				refreshnum--;
+				break;
+			default:
 				break;
 			}
 		}
@@ -656,6 +894,12 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 	if (refreshnum + upgradenum + installnum == 0)
 		goto installend;
 
+	/*
+	 * Perform any removals first.  Any superseded packages are highly
+	 * likely to conflict with incoming newer packages.
+	 */
+	do_pkg_remove(installhead);
+
 	/*
 	 * At this point we're performing installs.
 	 */
@@ -671,14 +915,15 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 	if (rebuild_required_by() != EXIT_SUCCESS)
 		rc = EXIT_FAILURE;
 
-	/* pure install, not called by pkgin_upgrade */
-	if (!upgrade)
-		(void)update_db(LOCAL_SUMMARY, pkgargs, 1);
+	(void)update_db(LOCAL_SUMMARY, 1);
 
 installend:
-
+	XFREE(todownload);
 	XFREE(toinstall);
+	XFREE(torefresh);
 	XFREE(toupgrade);
+	XFREE(toremove);
+	XFREE(tosupersede);
 	XFREE(unmet_reqs);
 	free_pkglist(&impacthead);
 	free_pkglist(&downloadhead);
@@ -688,6 +933,7 @@ installend:
 	 */
 	if (installhead != NULL)
 		free_pkglist(&installhead);
+	free(installhead);
 
 	return rc;
 }
@@ -695,73 +941,57 @@ installend:
 int
 pkgin_remove(char **pkgargs)
 {
-	int		deletenum = 0, exists, rc = EXIT_SUCCESS;
-	Plisthead	*pdphead, *removehead;
-	Pkglist		*pdp;
-	char   		*todelete = NULL, **ppkgargs, *pkgname, *ppkg;
+	Plistarray	*rmhead;
+	Plisthead	*removehead;
+	Pkglist		*lpkg, *p;
+	int		deletenum = 0, rc = EXIT_SUCCESS;
+	char   		*todelete = NULL, **arg;
 
-	pdphead = init_head();
-
-	if (SLIST_EMPTY(&l_plisthead))
+	if (is_empty_local_pkglist())
 		errx(EXIT_FAILURE, MSG_EMPTY_LOCAL_PKGLIST);
 
-	/* act on every package passed to the command line */
-	for (ppkgargs = pkgargs; *ppkgargs != NULL; ppkgargs++) {
+	rmhead = init_array(1);
 
-		if ((pkgname =
-			simple_pkg_match(&l_plisthead, *ppkgargs)) == NULL) {
-			printf(MSG_PKG_NOT_INSTALLED, *ppkgargs);
+	/*
+	 * For every package or pattern on the command line, find a matching
+	 * installed package and add to rmhead.
+	 */
+	for (arg = pkgargs; *arg != NULL; arg++) {
+		if ((lpkg = find_local_pkg(*arg, NULL)) == NULL) {
+			printf(MSG_PKG_NOT_INSTALLED, *arg);
 			rc = EXIT_FAILURE;
 			continue;
 		}
-		ppkg = xstrdup(pkgname);
-		trunc_str(ppkg, '-', STR_BACKWARD);
+		free(*arg);
+		*arg = xstrdup(lpkg->full);
 
-		/* record full reverse dependency list for package */
-		full_dep_tree(ppkg, LOCAL_REVERSE_DEPS, pdphead);
+		/*
+		 * Fetch full reverse dependencies for local package and add
+		 * them to rmhead too.
+		 */
+		get_depends_recursive(lpkg->full, rmhead, DEPENDS_REVERSE);
 
-		XFREE(ppkg);
+		/*
+		 * Add the package itself.
+		 */
+		p = malloc_pkglist();
+		p->lpkg = lpkg;
+		SLIST_INSERT_HEAD(&rmhead->head[0], p, next);
+	}
 
-		exists = 0;
-		/* check if package have already been recorded */
-		SLIST_FOREACH(pdp, pdphead, next) {
-			if (strncmp(pdp->depend, pkgname,
-					strlen(pdp->depend)) == 0) {
-				exists = 1;
-				break;
-			}
-		}
-
-		if (exists)
-			continue; /* next pkgarg */
-
-
-		/* add package itself */
-		pdp = malloc_pkglist();
-
-		pdp->depend = pkgname;
-
-		if (SLIST_EMPTY(pdphead))
-			/*
-			 * identify unique package,
-			 * don't cut it when ordering
-			 */
-			pdp->level = -1;
-		else
-			pdp->level = 0;
-
-		pdp->name = xstrdup(pdp->depend);
-		trunc_str(pdp->name, '-', STR_BACKWARD);
-
-		SLIST_INSERT_HEAD(pdphead, pdp, next);
-	} /* for pkgargs */
+	/*
+	 * Add ACTION_REMOVE to all entries.
+	 */
+	SLIST_FOREACH(p, &rmhead->head[0], next) {
+		p->action = ACTION_REMOVE;
+	}
 
 	/* order remove list */
-	removehead = order_remove(pdphead);
+	removehead = order_remove(&rmhead->head[0]);
 
-	SLIST_FOREACH(pdp, removehead, next) {
+	SLIST_FOREACH(p, removehead, next) {
 		deletenum++;
-		todelete = action_list(todelete, pdp->depend);
+		todelete = action_list(todelete, p->lpkg->full);
 	}
 
 	if (todelete != NULL) {
@@ -778,132 +1008,23 @@ pkgin_remove(char **pkgargs)
 			if (rebuild_required_by() != EXIT_SUCCESS)
 				rc = EXIT_FAILURE;
 
-			(void)update_db(LOCAL_SUMMARY, NULL, 1);
+			(void)update_db(LOCAL_SUMMARY, 1);
 		}
 	} else
 		printf(MSG_NO_PKGS_TO_DELETE);
 
 	free_pkglist(&removehead);
-	free_pkglist(&pdphead);
-
+	free_array(rmhead);
 	XFREE(todelete);
 
 	return rc;
 }
 
-/*
- * Find best match for a package to be upgraded.
- */
-static char *
-narrow_match(Pkglist *opkg)
-{
-	Pkglist	*pkglist;
-	char	*best_match;
-	int	refresh = 0;
-
-	/* for now, best match is old package itself */
-	best_match = xstrdup(opkg->full);
-
-	SLIST_FOREACH(pkglist, &r_plisthead, next) {
-		/* not the same pkgname, next */
-		if (pkgstrcmp(opkg->name, pkglist->name))
-			continue;
-
-		/*
-		 * Do not propose an upgrade if the PKGPATH does not match,
-		 * otherwise users who have requested a specific version of a
-		 * package for which there are usually multiple versions
-		 * available (e.g. nodejs or mysql) would always have it
-		 * replaced by the latest version.
-		 *
-		 * Note that this comparison does allow both PKGPATH values
-		 * to be NULL, for example with local self-built packages, in
-		 * which case we permit an upgrade.
-		 */
-		if (pkgstrcmp(opkg->pkgpath, pkglist->pkgpath))
-			continue;
-
-		/*
-		 * If the package version is identical, check if the BUILD_DATE
-		 * has changed.  If it has, we need to refresh the package as
-		 * it has been rebuilt, possibly against newer dependencies.
-		 */
-		if (pkgstrcmp(opkg->full, pkglist->full) == 0) {
-			if (pkgstrcmp(opkg->build_date, pkglist->build_date))
-				refresh = 1;
-			continue;
-		}
-
-		/* second package is greater */
-		if (version_check(best_match, pkglist->full) == 2) {
-			XFREE(best_match);
-			best_match = xstrdup(pkglist->full);
-		}
-	} /* SLIST_FOREACH remoteplisthead */
-
-	/* there was no upgrade candidate */
-	if (strcmp(best_match, opkg->full) == 0 && !refresh)
-		XFREE(best_match);
-
-	return best_match;
-}
-
-static char **
-record_upgrades(Plisthead *plisthead)
-{
-	Pkglist	*pkglist;
-	size_t	count = 0;
-	char	**pkgargs;
-
-	SLIST_FOREACH(pkglist, plisthead, next)
-		count++;
-
-	pkgargs = xmalloc((count + 2) * sizeof(char *));
-
-	count = 0;
-	SLIST_FOREACH(pkglist, plisthead, next) {
-		pkgargs[count] = narrow_match(pkglist);
-
-		if (pkgargs[count] == NULL)
-			continue;
-
-		count++;
-	}
-	pkgargs[count] = NULL;
-
-	return pkgargs;
-}
-
 int
 pkgin_upgrade(int do_inst)
 {
-	Plistnumbered	*keeplisthead;
-	char		**pkgargs;
-	int		rc;
-
-	/* record keepable packages */
-	if ((keeplisthead = rec_pkglist(KEEP_LOCAL_PKGS)) == NULL)
-		errx(EXIT_FAILURE, MSG_EMPTY_KEEP_LIST);
-
-	/* upgrade all packages, not only keepables */
-	if (SLIST_EMPTY(&l_plisthead))
+	if (is_empty_local_pkglist())
 		errx(EXIT_FAILURE, MSG_EMPTY_LOCAL_PKGLIST);
 
-	pkgargs = record_upgrades(&l_plisthead);
-
-	rc = pkgin_install(pkgargs, do_inst, 1);
-
-	free_list(pkgargs);
-
-	/* Record keep list */
-	if (do_inst) {
-		pkgargs = record_upgrades(keeplisthead->P_Plisthead);
-		(void)update_db(LOCAL_SUMMARY, pkgargs, 1);
-		free_list(pkgargs);
-	}
-
-	free_pkglist(&keeplisthead->P_Plisthead);
-	free(keeplisthead);
-
-	return rc;
+	return pkgin_install(NULL, do_inst, 1);
 }
