@@ -30,6 +30,51 @@
 #include <sqlite3.h>
 #include "pkgin.h"
 
+/*
+ * SQLite callback for LOCAL_CONFLICTS etc, record a pattern and optional
+ * PKGBASE to a Plistarray.
+ *
+ * argv0: pattern
+ * argv1: pkgbase, may be NULL if it cannot be determined from pattern
+ */
+static int
+record_pattern_to_array(void *param, int argc, char **argv, char **colname)
+{
+	Plistarray *depends = (Plistarray *)param;
+	Pkglist *d;
+	int slot;
+
+	if (argv == NULL)
+		return PDB_ERR;
+
+	d = malloc_pkglist();
+	d->patterns = xmalloc(2 * sizeof(char *));
+	d->patterns[0] = xstrdup(argv[0]);
+	d->patterns[1] = NULL;
+	d->patcount = 1;
+
+	/*
+	 * XXX: default slot if no pkgbase available, should we allocate one
+	 * outside of the normal range for these?
+	 */
+	slot = 0;
+
+	if (argv[1]) {
+		d->name = xstrdup(argv[1]);
+		slot = pkg_hash_entry(d->name, depends->size);
+	}
+
+	SLIST_INSERT_HEAD(&depends->head[slot], d, next);
+
+	return PDB_OK;
+}
+
+void
+get_conflicts(Plistarray *conflicts)
+{
+	pkgindb_doquery(LOCAL_CONFLICTS, record_pattern_to_array, conflicts);
+}
+
 /* find required files (REQUIRES) from PROVIDES or filename */
 int
 pkg_met_reqs(Plisthead *impacthead)
@@ -48,9 +93,16 @@ pkg_met_reqs(Plisthead *impacthead)
 
 	/* first, parse impact list */
 	SLIST_FOREACH(pimpact, impacthead, next) {
+		/*
+		 * Skip entries that only operate on local packages (e.g.
+		 * ACTION_SUPERSEDED).
+		 */
+		if (pimpact->rpkg == NULL)
+			continue;
+
 		/* retreive requires list for package */
-		if ((requireshead = rec_pkglist(GET_REQUIRES_QUERY,
-					pimpact->full)) == NULL)
+		if ((requireshead = rec_pkglist(REMOTE_REQUIRES,
+					pimpact->rpkg->full)) == NULL)
 			/* empty requires list (very unlikely) */
 			continue;
 
@@ -66,13 +118,13 @@ pkg_met_reqs(Plisthead *impacthead)
 				    strlen(PREFIX) - 1)) != 0) {
 				if (stat(requires->full, &sb) < 0) {
 					printf(MSG_REQT_NOT_PRESENT,
-						requires->full, pimpact->full);
+						requires->full, pimpact->rpkg->full);
 
 					/*
 					 * mark as DONOTHING,
 					 * requirement missing
 					 */
-					pimpact->action = UNMET_REQ;
+					pimpact->action = ACTION_UNMET_REQ;
 
 					met_reqs = 0;
 				}
@@ -110,7 +162,7 @@ pkg_met_reqs(Plisthead *impacthead)
 				/* re-parse impact list to retreive PROVIDES */
 				SLIST_FOREACH(impactprov, impacthead, next) {
 					if ((r_provideshead =
-						rec_pkglist(GET_PROVIDES_QUERY,
+						rec_pkglist(REMOTE_PROVIDES,
 						impactprov->full)) == NULL)
 						continue;
 
@@ -172,44 +224,70 @@ pkg_met_reqs(Plisthead *impacthead)
 	return met_reqs;
 }
 
-/* check for conflicts and if needed files are present */
-int
-pkg_has_conflicts(Pkglist *pimpact)
+/*
+ * A remote package has been identified as matching a local conflict.  Retrieve
+ * the local package and print the conflict.
+ */
+static void
+print_pkg_conflict(const char *pattern, const char *pkgname)
 {
-	int			has_conflicts = 0;
-	char		*conflict_pkg, query[BUFSIZ];
-	Pkglist		*conflicts; /* SLIST conflicts pointer */
-	Plistnumbered	*conflictshead;
+	char query[BUFSIZ];
+	char *conflict;
 
-	/* conflicts list */
-	if ((conflictshead = rec_pkglist(LOCAL_CONFLICTS)) == NULL)
+	conflict = xmalloc(BUFSIZ * sizeof(char));
+
+	sqlite3_snprintf(BUFSIZ, query, REMOTE_CONFLICTS, pattern);
+	if (pkgindb_doquery(query, pdb_get_value, conflict) == PDB_OK)
+		printf(MSG_CONFLICT_PKG, pkgname, conflict);
+
+	XFREE(conflict);
+}
+
+/*
+ * Check if an incoming remote package conflicts with any local packages.
+ */
+int
+pkg_has_conflicts(Pkglist *pkg, Plistarray *conflicts)
+{
+	Pkglist *p;
+	int i, slot;
+
+	if (is_empty_plistarray(conflicts))
 		return 0;
 
-	/* check conflicts */
-	SLIST_FOREACH(conflicts, conflictshead->P_Plisthead, next) {
-		if (pkg_match(conflicts->full, pimpact->full)) {
+	slot = pkg_hash_entry(pkg->rpkg->name, conflicts->size);
+	SLIST_FOREACH(p, &conflicts->head[slot], next) {
+		for (i = 0; i < p->patcount; i++) {
+			if (!pkg_match(p->patterns[i], pkg->rpkg->full))
+				continue;
 
-			/* got a conflict, retrieve conflicting local package */
-			sqlite3_snprintf(BUFSIZ, query,
-			    GET_CONFLICT_QUERY, conflicts->full);
+			/*
+			 * Incoming remote package matches a local CONFLICTS
+			 * entry, print the conflict message and return fail.
+			 */
+			print_pkg_conflict(p->patterns[i], pkg->rpkg->full);
+			return 1;
+		}
+	}
 
-			conflict_pkg = xmalloc(BUFSIZ * sizeof(char));
-			if (pkgindb_doquery(query,
-					pdb_get_value, conflict_pkg) == PDB_OK)
+	/*
+	 * We also need to check slot 0 if not already checked as that's where
+	 * any CONFLICTS that have a complicated pattern and no pkgbase are
+	 * stored.
+	 */
+	if (slot == 0)
+		return 0;
 
-				printf(MSG_CONFLICT_PKG,
-					pimpact->full, conflict_pkg);
+	SLIST_FOREACH(p, &conflicts->head[0], next) {
+		for (i = 0; i < p->patcount; i++) {
+			if (!pkg_match(p->patterns[i], pkg->rpkg->full))
+				continue;
+			print_pkg_conflict(p->patterns[i], pkg->rpkg->full);
+			return 1;
+		}
+	}
 
-			XFREE(conflict_pkg);
-
-			has_conflicts = 1;
-		} /* match conflict */
-	} /* SLIST_FOREACH conflicts */
-
-	free_pkglist(&conflictshead->P_Plisthead);
-	free(conflictshead);
-
-	return has_conflicts;
+	return 0;
 }
 
 void
@@ -224,7 +302,7 @@ show_prov_req(const char *query, const char *pkgname)
 	if ((fullpkgname = unique_pkg(pkgname, REMOTE_PKG)) == NULL)
 		errx(EXIT_FAILURE, MSG_PKG_NOT_AVAIL, pkgname);
 
-	say = ( query == GET_PROVIDES_QUERY ) ? out[0] : out[1];
+	say = (query == REMOTE_PROVIDES) ? out[0] : out[1];
 
 	if ((plisthead = rec_pkglist(query, fullpkgname)) == NULL) {
 		printf(MSG_NO_PROV_REQ, say, fullpkgname);
@@ -235,5 +313,7 @@ show_prov_req(const char *query, const char *pkgname)
 	SLIST_FOREACH(plist, plisthead->P_Plisthead, next)
 		printf("\t%s\n", plist->full);
 
+	free_pkglist(&plisthead->P_Plisthead);
+	free(plisthead);
 	XFREE(fullpkgname);
 }
