@@ -27,108 +27,97 @@
  * SUCH DAMAGE.
  */
 
-const char DROP_LOCAL_TABLES[] =
-	"DROP TABLE IF EXISTS LOCAL_DEPS;"
-	"DROP TABLE IF EXISTS LOCAL_PKG;"
-	"DROP TABLE IF EXISTS LOCAL_CONFLICTS;"
-	"DROP TABLE IF EXISTS LOCAL_REQUIRES;"
-	"DROP TABLE IF EXISTS LOCAL_PROVIDES;";
-
-const char DROP_REMOTE_TABLES[] =
-	"DROP TABLE IF EXISTS REMOTE_DEPS;"
-	"DROP TABLE IF EXISTS REMOTE_PKG;"
-	"DROP TABLE IF EXISTS REMOTE_CONFLICTS;"
-	"DROP TABLE IF EXISTS REMOTE_REQUIRES;"
-	"DROP TABLE IF EXISTS REMOTE_PROVIDES;";
+/*
+ * This query checks the compatibility of the current database, and should be
+ * one that either completes or fails due to an SQL error based on the most
+ * recent schema change.  Returned rows are ignored, so choose a query that
+ * runs quickly.
+ */
+const char CHECK_DB_LATEST[] =
+	"SELECT pkgbase "
+	"  FROM local_conflicts "
+	" LIMIT 1;";
 
 const char DELETE_LOCAL[] =
-	"DELETE FROM LOCAL_DEPS;"
 	"DELETE FROM LOCAL_PKG;"
 	"DELETE FROM LOCAL_CONFLICTS;"
+	"DELETE FROM LOCAL_DEPENDS;"
+	"DELETE FROM LOCAL_PROVIDES;"
 	"DELETE FROM LOCAL_REQUIRES;"
-	"DELETE FROM LOCAL_PROVIDES;";
+	"DELETE FROM LOCAL_REQUIRED_BY;";
 
 const char DELETE_REMOTE[] =
-	"DELETE FROM %s WHERE %s_ID IN "
-	"(SELECT %s.%s_ID FROM REMOTE_PKG, %s "
-	"WHERE REMOTE_PKG.REPOSITORY GLOB %Q || '*' AND "
-	"REMOTE_PKG.PKG_ID = %s.PKG_ID);";
+	"DELETE FROM %s "
+	" WHERE pkg_id IN "
+	"    (SELECT pkg_id "
+	"       FROM remote_pkg "
+	"      WHERE repository GLOB %Q || '*' "
+	"    );";
 
 const char DELETE_REMOTE_PKG_REPO[] =
 	"DELETE FROM REMOTE_PKG WHERE REPOSITORY = %Q;";
 
-const char DIRECT_DEPS[] = /* prefer higher version */
-	"SELECT REMOTE_DEPS_DEWEY, REMOTE_DEPS_PKGNAME "
-	"FROM REMOTE_DEPS WHERE PKG_ID = "
-	"(SELECT PKG_ID FROM REMOTE_PKG WHERE PKGNAME = %Q "
-	"ORDER BY FULLPKGNAME DESC LIMIT 1);";
+const char LOCAL_DIRECT_DEPENDS[] =
+	"SELECT pattern, pkgbase "
+	"  FROM local_depends, local_pkg "
+	" WHERE fullpkgname = %Q "
+	"   AND local_depends.pkg_id = local_pkg.pkg_id;";
 
-const char LOCAL_DIRECT_DEPS[] =
-	"SELECT LOCAL_DEPS_DEWEY, LOCAL_DEPS_PKGNAME "
-	"FROM LOCAL_DEPS WHERE PKG_ID = "
-	"(SELECT PKG_ID FROM LOCAL_PKG WHERE PKGNAME = %Q "
-	"ORDER BY FULLPKGNAME DESC LIMIT 1);";
+const char REMOTE_DIRECT_DEPENDS[] =
+	"SELECT pattern, pkgbase "
+	"  FROM remote_depends, remote_pkg "
+	" WHERE fullpkgname = %Q "
+	"   AND remote_depends.pkg_id = remote_pkg.pkg_id;";
 
-const char EXACT_DIRECT_DEPS[] =
-	"SELECT REMOTE_DEPS.REMOTE_DEPS_DEWEY, REMOTE_DEPS.REMOTE_DEPS_PKGNAME "
-	"FROM REMOTE_DEPS,REMOTE_PKG "
-	"WHERE REMOTE_PKG.FULLPKGNAME = %Q "
-	"AND REMOTE_DEPS.PKG_ID = REMOTE_PKG.PKG_ID;";
-
-const char LOCAL_REVERSE_DEPS[] =
-	"SELECT LOCAL_PKG.FULLPKGNAME, LOCAL_PKG.PKGNAME, LOCAL_PKG.PKG_KEEP "
-	"FROM LOCAL_PKG, LOCAL_DEPS "
-	"WHERE LOCAL_DEPS.LOCAL_DEPS_PKGNAME  = %Q "
-	"AND LOCAL_PKG.PKG_ID = LOCAL_DEPS.PKG_ID;";
-
-const char REMOTE_REVERSE_DEPS[] =
-	"SELECT REMOTE_PKG.FULLPKGNAME, REMOTE_PKG.PKGNAME "
-	"FROM REMOTE_PKG, REMOTE_DEPS, LOCAL_DEPS "
-	"WHERE LOCAL_DEPS.LOCAL_DEPS_PKGNAME  = %Q "
-	"AND REMOTE_PKG.PKG_ID = REMOTE_DEPS.PKG_ID;";
+const char LOCAL_REVERSE_DEPENDS[] =
+	"SELECT required_by, local_pkg.pkgname, local_pkg.pkg_keep "
+	"  FROM local_pkg "
+	"  LEFT JOIN local_required_by "
+	"    ON local_pkg.fullpkgname = local_required_by.required_by "
+	" WHERE local_required_by.pkgname = %Q;";
 
 const char LOCAL_CONFLICTS[] =
-	"SELECT LOCAL_CONFLICTS_PKGNAME FROM LOCAL_CONFLICTS;";
-
-const char GET_CONFLICT_QUERY[] =
-	"SELECT LOCAL_PKG.FULLPKGNAME FROM LOCAL_CONFLICTS,LOCAL_PKG "
-	"WHERE LOCAL_CONFLICTS.LOCAL_CONFLICTS_PKGNAME = %Q "
-	"AND LOCAL_CONFLICTS.PKG_ID = LOCAL_PKG.PKG_ID;";
-
-/* naming dirt, this is not really a PKGNAME, but this shortcut permits
- * the use of a generic function in summary.c (child_table)
- */
-const char GET_REQUIRES_QUERY[] =
-	"SELECT REMOTE_REQUIRES.REMOTE_REQUIRES_PKGNAME "
-	"FROM REMOTE_REQUIRES,REMOTE_PKG "
-	"WHERE REMOTE_PKG.FULLPKGNAME = %Q "
-	"AND REMOTE_REQUIRES.PKG_ID = REMOTE_PKG.PKG_ID;";
-
-const char GET_PROVIDES_QUERY[] =
-	"SELECT REMOTE_PROVIDES.REMOTE_PROVIDES_PKGNAME "
-	"FROM REMOTE_PROVIDES,REMOTE_PKG "
-	"WHERE REMOTE_PKG.FULLPKGNAME = %Q "
-	"AND REMOTE_PROVIDES.PKG_ID = REMOTE_PKG.PKG_ID;";
+	"SELECT DISTINCT pattern, pkgbase "
+	"  FROM local_conflicts;";
 
 const char LOCAL_PROVIDES[] =
-	"SELECT LOCAL_PROVIDES_PKGNAME FROM LOCAL_PROVIDES;";
+	"SELECT filename "
+	"  FROM local_provides;";
+
+const char REMOTE_CONFLICTS[] =
+	"SELECT local_pkg.fullpkgname "
+	"  FROM local_conflicts, local_pkg "
+	" WHERE local_conflicts.pattern = %Q "
+	"   AND local_conflicts.pkg_id = local_pkg.pkg_id;";
+
+const char REMOTE_PROVIDES[] =
+	"SELECT filename "
+	"  FROM remote_provides, remote_pkg "
+	" WHERE fullpkgname = %Q "
+	"   AND remote_provides.pkg_id = remote_pkg.pkg_id;";
+
+const char REMOTE_REQUIRES[] =
+	"SELECT filename "
+	"  FROM remote_requires, remote_pkg "
+	" WHERE fullpkgname = %Q "
+	"   AND remote_requires.pkg_id = remote_pkg.pkg_id;";
+
+const char REMOTE_SUPERSEDES[] =
+	"SELECT pattern, pkgbase "
+	"  FROM remote_supersedes "
+	"  LEFT JOIN remote_pkg "
+	"    ON remote_supersedes.pkg_id = remote_pkg.pkg_id "
+	" WHERE fullpkgname = %Q;";
 
 const char KEEP_PKG[] =
 	"UPDATE LOCAL_PKG SET PKG_KEEP = 1 WHERE PKGNAME = %Q;";
 const char UNKEEP_PKG[] =
 	"UPDATE LOCAL_PKG SET PKG_KEEP = NULL WHERE PKGNAME = %Q;";
 
-const char INSERT_REMOTE_PKG[] =
-	"INSERT INTO REMOTE_PKG (PKG_ID,FULLPKGNAME,PKGNAME,PKGVERS,BUILD_DATE,"
-	"COMMENT,LICENSE,PKGTOOLS_VERSION,HOMEPAGE,OS_VERSION,PKGPATH,"
-	"PKG_OPTIONS,CATEGORIES,SIZE_PKG,FILE_SIZE,OPSYS,REPOSITORY) "
-	"VALUES (%d,%Q,%Q,%Q,%Q,%Q,%Q,%Q,%Q"
-	"%Q,%Q,%Q,%Q,%Q,%Q,%Q,%Q);";
-
 /* for upgrades, prefer higher versions to be at the top of SLIST */
 const char LOCAL_PKGS_QUERY_ASC[] =
 	"SELECT FULLPKGNAME,PKGNAME,PKGVERS,BUILD_DATE,"
-	"COMMENT,FILE_SIZE,SIZE_PKG,CATEGORIES,PKGPATH "
+	"COMMENT,FILE_SIZE,SIZE_PKG,CATEGORIES,PKGPATH,PKG_KEEP "
 	"FROM LOCAL_PKG "
 	"ORDER BY FULLPKGNAME ASC;";
 
@@ -154,12 +143,16 @@ const char REMOTE_PKGS_QUERY_DESC[] =
 	"ORDER BY FULLPKGNAME DESC;";
 
 const char NOKEEP_LOCAL_PKGS[] =
-	"SELECT FULLPKGNAME,PKGNAME,PKGPATH,COMMENT "
-	"FROM LOCAL_PKG WHERE PKG_KEEP IS NULL;";
+	"SELECT fullpkgname,pkgname,pkgpath,comment "
+	"  FROM local_pkg "
+	" WHERE pkg_keep IS NULL "
+	" ORDER BY fullpkgname DESC;";
 
 const char KEEP_LOCAL_PKGS[] =
-	"SELECT FULLPKGNAME,PKGNAME,PKGPATH,COMMENT "
-	"FROM LOCAL_PKG WHERE PKG_KEEP IS NOT NULL;";
+	"SELECT fullpkgname,pkgname,pkgpath,comment "
+	"  FROM local_pkg "
+	" WHERE pkg_keep IS NOT NULL"
+	" ORDER BY fullpkgname DESC;";
 
 const char PKG_URL[] =
 	"SELECT REPOSITORY FROM REMOTE_PKG WHERE FULLPKGNAME = %Q;";
@@ -182,11 +175,23 @@ const char UPDATE_REPO_MTIME[] =
 const char DELETE_REPO_URL[] =
 	"DELETE FROM REPOS WHERE REPO_URL = %Q;";
 
-const char INSERT_SINGLE_VALUE[] =
-	"INSERT INTO %s (PKG_ID, %s_PKGNAME) VALUES (%d,%Q);";
+const char INSERT_CONFLICTS[] =
+	"INSERT INTO %s (pkg_id, pattern, pkgbase) VALUES (%d, %Q, %Q);";
 
-const char INSERT_DEPENDS_VALUES[] =
-	"INSERT INTO %s (PKG_ID, %s_PKGNAME, %s_DEWEY) VALUES (%d,%Q,%Q);";
+const char INSERT_DEPENDS[] =
+	"INSERT INTO %s (pkg_id, pattern, pkgbase) VALUES (%d, %Q, %Q);";
+
+const char INSERT_PROVIDES[] =
+	"INSERT INTO %s (PKG_ID, FILENAME) VALUES (%d, %Q);";
+
+const char INSERT_REQUIRES[] =
+	"INSERT INTO %s (PKG_ID, FILENAME) VALUES (%d, %Q);";
+
+const char INSERT_SUPERSEDES[] =
+	"INSERT INTO %s (pkg_id, pattern, pkgbase) VALUES (%d, %Q, %Q);";
+
+const char INSERT_REQUIRED_BY[] =
+	"INSERT INTO LOCAL_REQUIRED_BY (PKGNAME, REQUIRED_BY) VALUES (%Q, %Q);";
 
 const char UNIQUE_PKG[] =
 	"SELECT FULLPKGNAME, PKGVERS FROM %s WHERE PKGNAME = %Q;";
@@ -202,10 +207,6 @@ const char EXPORT_KEEP_LIST[] =
 const char GET_PKGNAME_BY_PKGPATH[] =
 	"SELECT FULLPKGNAME FROM REMOTE_PKG WHERE PKGPATH = %Q;";
 
-const char GET_ORPHAN_PACKAGES[] =
-	"SELECT FULLPKGNAME FROM LOCAL_PKG WHERE PKG_KEEP IS NULL AND "
-	"PKGNAME NOT IN (SELECT LOCAL_DEPS_PKGNAME FROM LOCAL_DEPS);";
-
 const char SHOW_ALL_CATEGORIES[] =
 	"SELECT DISTINCT CATEGORIES FROM REMOTE_PKG WHERE "
 	"CATEGORIES NOT LIKE '%% %%' ORDER BY CATEGORIES DESC;";
