@@ -61,32 +61,37 @@ pdb_rec_list(void *param, int argc, char **argv, char **colname)
 		if (argv[i] == NULL)
 			continue;
 
-		if (strcmp(colname[i], "PKGNAME") == 0)
+		if (strcmp(colname[i], "PKGNAME") == 0) {
 			plist->name = xstrdup(argv[i]);
-		if (strcmp(colname[i], "PKGVERS") == 0)
+			continue;
+		}
+		if (strcmp(colname[i], "PKGVERS") == 0) {
 			plist->version = xstrdup(argv[i]);
-		if (strcmp(colname[i], "BUILD_DATE") == 0)
+			continue;
+		}
+		if (strcmp(colname[i], "BUILD_DATE") == 0) {
 			plist->build_date = xstrdup(argv[i]);
-		if (strcmp(colname[i], "COMMENT") == 0)
+			continue;
+		}
+		if (strcmp(colname[i], "COMMENT") == 0) {
 			plist->comment = xstrdup(argv[i]);
-		if (strcmp(colname[i], "PKGPATH") == 0)
+			continue;
+		}
+		if (strcmp(colname[i], "PKGPATH") == 0) {
 			plist->pkgpath = xstrdup(argv[i]);
-		if (strcmp(colname[i], "CATEGORIES") == 0)
+			continue;
+		}
+		if (strcmp(colname[i], "CATEGORIES") == 0) {
 			plist->category = xstrdup(argv[i]);
-		if (strcmp(colname[i], "FILE_SIZE") == 0)
+			continue;
+		}
+		if (strcmp(colname[i], "FILE_SIZE") == 0) {
 			plist->file_size = strtol(argv[i], (char **)NULL, 10);
-		if (strcmp(colname[i], "SIZE_PKG") == 0)
+			continue;
+		}
+		if (strcmp(colname[i], "SIZE_PKG") == 0) {
 			plist->size_pkg = strtol(argv[i], (char **)NULL, 10);
-	}
-
-	/*
-	 * Verify FILE_SIZE exists for remote packages.
-	 */
-	if (plisthead->P_type == 1) {
-		if (plist->file_size == 0) {
-			fprintf(stderr, MSG_BAD_FILE_SIZE, plist->full);
-			free_pkglist_entry(&plist);
-			return PDB_ERR;
+			continue;
 		}
 	}
 
@@ -95,63 +100,3 @@ pdb_rec_list(void *param, int argc, char **argv, char **colname)
 
 	return PDB_OK;
 }
-
-/* dewey for direct deps, full pkgname for reverse deps*/
-#define DEPS_FULLPKG		argv[0]
-/* pkgname for direct deps and reverse deps */
-#define DEPS_PKGNAME		argv[1]
-#define PKG_KEEP		argv[2]
-/**
- * sqlite callback
- * DIRECT_DEPS or REVERSE_DEPS result, feeds a Pkglist SLIST
- * Plisthead is the head of Pkglist
- */
-int
-pdb_rec_depends(void *param, int argc, char **argv, char **colname)
-{
-	Pkglist		*deptree, *pdp, *pkg_map;
-	Plisthead	*pdphead = (Plisthead *)param, *plisthead;
-
-	if (argv == NULL)
-		return PDB_ERR;
-
-	/* check if dependency is already recorded, do not insert on list  */
-	SLIST_FOREACH(pdp, pdphead, next)
-		if (strcmp(DEPS_PKGNAME, pdp->name) == 0) {
-			TRACE(" < dependency %s already recorded\n", pdp->name);
-			/* proceed to next result */
-			return PDB_OK;
-		}
-
-	deptree = malloc_pkglist();
-	deptree->depend = xstrdup(DEPS_FULLPKG);
-
-	/* unresolved pkgname because of complex dependency glob */
-	if (non_trivial_glob(DEPS_FULLPKG)) {
-		/* check wether we're getting local or remote dependencies */
-		if (strncmp(colname[0], "LOCAL_", 6) == 0)
-			plisthead = &l_plisthead;
-		else
-			plisthead = &r_plisthead;
-
-		/* map corresponding pkgname */
-		if ((pkg_map = find_pkg_match(plisthead, deptree->depend)) != NULL)
-			deptree->name = xstrdup(pkg_map->name);
-		else
-			/* some dependencies just don't match anything */
-			deptree->name = xstrdup(DEPS_PKGNAME);
-	} else
-		/* case handled by get_pkgname_from_depend() */
-		deptree->name = xstrdup(DEPS_PKGNAME);
-
-	deptree->level = 0;
-	/* used in LOCAL_REVERSE_DEPS / autoremove.c */
-	if (argc > 2 && PKG_KEEP != NULL)
-		deptree->keep = 1;
-	else
-		deptree->keep = 0;
-
-	SLIST_INSERT_HEAD(pdphead, deptree, next);
-
-	return PDB_OK;
-}
